#include "ac.h"
#include "ac_stdio.h"
#include "ac_url.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

AC_FILE *ac_fopen(const char *url, const char *mode) {
	AC_FILE *f;
	char *localurl;
	if (!mode) return NULL;
	
	if (ac_url_islocal(url)) {
		localurl = ac_url_makelocal(url);
		if (!localurl) {
			return NULL;
		}
		AC_DEBUG("opening %s in local mode", localurl);
		f = malloc(sizeof(AC_FILE));
		if (!f) return NULL;
		f->flags = AC_FILE_FLAG_LOCAL;
		f->localfile = fopen(localurl, mode);
		free(localurl);
		f->state = AC_STATE_BOUND;
		if (!f->localfile) {
			free(f);
			return NULL;
		} else {
			return f;
		}
	} else {
		if (mode[0] != 'r' && mode[0] != 'R') return NULL;
		f = malloc(sizeof(AC_FILE));
		if (!f) return NULL;
		f->flags = 0;
		f->offset = 0;
		f->curl = curl_easy_init();
		//curl_easy_setopt(f->curl, CURLOPT_VERBOSE, 1);
		curl_easy_setopt(f->curl, CURLOPT_URL, url);
		f->state = AC_STATE_BOUND;
		return f;
	}
}

static size_t ac_fread_callback(void *ptr, size_t size, size_t nmemb, void *stream) {
	AC_FILE *f;
	long left;
	f = (AC_FILE *)stream;
	left = f->toread;
	/* FIXME: ac_fwrite instead of memcpy */
	if (left > 0 && size * nmemb < left) {
		memcpy(f->curbuf + f->curoffset, ptr, size * nmemb);
		f->toread -= size * nmemb;
		f->offset += size * nmemb;
		f->curoffset += size * nmemb;
		return size * nmemb;
	} else {
		memcpy(f->curbuf + f->curoffset, ptr, left);
		f->toread = 0;
		f->offset += size * nmemb;
		f->curoffset += left;
		return left;
	}
}

size_t ac_fread(void *ptr, size_t size, size_t nelem, AC_FILE *stream) {
	char buf[256];
	int oldoffs;
	if (!stream || stream->state == AC_STATE_UNBOUND) return 0;
	if (stream->flags & AC_FILE_FLAG_LOCAL) {
		return fread(ptr, size, nelem, stream->localfile);
	}
	oldoffs = stream->offset;
	stream->curbuf = ptr;
	stream->toread = size * nelem;
	stream->curoffset = 0;
	curl_easy_setopt(stream->curl, CURLOPT_WRITEDATA, stream);
	curl_easy_setopt(stream->curl, CURLOPT_WRITEFUNCTION, ac_fread_callback);
	sprintf(buf, "%ld-%ld", stream->offset, stream->offset + size * nelem - 1);
	curl_easy_setopt(stream->curl, CURLOPT_RANGE, buf);
	curl_easy_perform(stream->curl);
	return stream->offset - oldoffs;
}

static size_t ac_fread_complete_callback(void *ptr, size_t size, size_t nmemb, void *stream) {
	AC_FILE *f;
	
	f = (AC_FILE *)stream;
	if (size * nmemb > f->toread - f->curoffset) {
		/* grow the buffer */
		while (size * nmemb > f->toread - f->curoffset) {
			f->toread *= 2;
		}
		f->curbuf = (char *)realloc(f->curbuf, f->toread);
	}
	memcpy(f->curbuf + f->curoffset, ptr, size * nmemb);
	f->curoffset += size * nmemb;
	return size * nmemb;
}

size_t ac_fread_complete(void **ptr, AC_FILE *stream) {
	char *buf;
	int size, actsize, read;
	int oldoffs = stream->offset;
	
	if (!stream || stream->state == AC_STATE_UNBOUND) return 0;
	if (stream->flags & AC_FILE_FLAG_LOCAL) {
		/* Do a complete read in local mode */
		size = 2048;
		actsize = 0;
		buf = malloc(sizeof(char) * size);
		while (read = fread(buf + actsize, 1, size - actsize, stream->localfile)) {
			actsize += read;
			if (read == size - actsize) {
				/* The buffer has to grow */
				size *= 2;
				buf = (char *)realloc(buf, size);
			}
		}
		*ptr = buf;
		return actsize;
	}
	stream->curbuf = (char *)malloc(sizeof(char) * 2048);
	stream->toread = 2048;
	stream->curoffset = 0;
	curl_easy_setopt(stream->curl, CURLOPT_WRITEDATA, stream);
	curl_easy_setopt(stream->curl, CURLOPT_WRITEFUNCTION, ac_fread_complete_callback);	
	curl_easy_perform(stream->curl);
	*ptr = stream->curbuf;
	stream->offset = stream->curoffset;
	return stream->offset - oldoffs;
}

int ac_fclose(AC_FILE *stream) {
	FILE *f;
	if (!stream || stream->state == AC_STATE_UNBOUND) return 1;
	stream->state = AC_STATE_UNBOUND;
	if (stream->flags & AC_FILE_FLAG_LOCAL) {
		f = stream->localfile;
		free(stream);
		return fclose(f);
	}
	curl_easy_cleanup(stream->curl);
	free(stream);
	return 0;
}

char *ac_fgets(char *s, int n, AC_FILE *stream) {
	int i;
	size_t read;
	char *buf;
	if (!stream || stream->state == AC_STATE_UNBOUND) return NULL;
	if (stream->flags & AC_FILE_FLAG_LOCAL) {
		return fgets(s, n, stream->localfile);
	}
	buf = malloc(sizeof(char) * (n - 1));
	read = ac_fread(buf, sizeof(char), n - 1, stream);
	if (read == 0) {
		free(buf);
		return NULL;
	}
	for (i = 0; i < read; i++) {
		s[i] = buf[i];
		if (buf[i] == '\0' || buf[i] == '\n') {
			break;
		} 
	}
	if (buf[i] != '\0') {
		s[i + 1] = '\0';
	}
	ac_fseek(stream, -read + i + 1, SEEK_CUR);
	free(buf);
	return s;
}

int ac_fseek(AC_FILE *stream, long offset, int mode) {
	if (!stream || stream->state == AC_STATE_UNBOUND) return 1;
	if (stream->flags & AC_FILE_FLAG_LOCAL) {
		return fseek(stream->localfile, offset, mode);
	}
	switch (mode) {
		case SEEK_SET:
			stream->offset = offset;
			break;
		case SEEK_END: /* FIXME: how do we know the end of the file? */
			return 1;
			break;
		case SEEK_CUR:
			stream->offset += offset;
			break;
	}
	return 0;
}

long ac_ftell(AC_FILE *stream) {
	if (!stream || stream->state == AC_STATE_UNBOUND) return -1;
	if (stream->flags & AC_FILE_FLAG_LOCAL) {
		return ftell(stream->localfile);
	}
	return stream->offset;
}
